"""Prometheus metrics for the tokenizer service."""
# TODO: disable use in development environments to avoid log spam.

from prometheus_client import CollectorRegistry, Counter, Histogram


class Metrics:
    """Prometheus metrics for the tokenizer service."""

    _instance = None
    _registry = CollectorRegistry()

    def __new__(cls):
        """Singleton pattern that allows arbitrary arguments."""
        if cls._instance is None:
            cls._instance = super(Metrics, cls).__new__(cls)
            cls._initialize_metrics(cls._instance)
        return cls._instance

    @classmethod
    def _initialize_metrics(cls, instance):
        """Initialize metrics."""
        instance.FUNCTION_LATENCY = Histogram(
            "tokenizer_function_latency_seconds",
            "Function latency (seconds)",
            buckets=[
                0.001,
                0.003,
                0.005,
                0.01,
                0.025,
                0.05,
                0.075,
                0.1,
                0.25,
                0.5,
                0.75,
                1.0,
                2.5,
                5.0,
                7.5,
                10.0,
                15.0,
                20.0,
                25.0,
                30.0,
                float("inf"),
            ],
            labelnames=["function"],
            registry=cls._registry,
        )

        instance.HERMES_FAILURE_COUNT = Counter(
            "tokenizer_hermes_failure_count",
            "Number of failures when calling Hermes",
            labelnames=["reason"],
            registry=cls._registry,
        )

        instance.HERMES_TEMPLATE_NAME = Counter(
            "tokenizer_hermes_template_name",
            "Count the number of times a specific template name is used.",
            labelnames=["template_name", "source"],
            registry=cls._registry,
        )

        instance.TOKEN_METRICS = Histogram(
            "tokenizer_token_metrics",
            "Metrics reflecting how tokens are handled in tokenizer.",
            buckets=[
                1000,
                1500,
                2000,
                2500,
                3000,
                3500,
                4000,
                5000,
                6000,
                7000,
                8000,
                9000,
                10000,
                15000,
                20000,
                25000,
                30000,
                35000,
                40000,
                45000,
                50000,
                60000,
                70000,
                80000,
                90000,
                100000,
                250000,
                500000,
                750000,
                1000000,
                float("inf"),
            ],
            labelnames=["metric_type"],
            registry=cls._registry,
        )

        instance.MESSAGE_METRICS = Histogram(
            "tokenizer_message_metrics",
            "Metrics reflecting how messages are handled in tokenizer.",
            buckets=[
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                15,
                20,
                25,
                30,
                35,
                40,
                45,
                50,
                60,
                70,
                80,
                90,
                100,
                150,
                200,
                250,
                300,
                350,
                400,
                450,
                500,
                600,
                700,
                800,
                900,
                1000,
                1250,
                1500,
                1750,
                2000,
                2500,
                3000,
                3500,
                4000,
                4500,
                5000,
                float("inf"),
            ],
            labelnames=["metric_type"],
            registry=cls._registry,
        )



"""Hermes templating engine for structured prefix generation."""

# TODO: import ansible core filters.
# TODO: Move this out of the hermes wheel and back into the tokenizer core.

import re
from logging import LoggerAdapter
from typing import Union

from constants import DEFAULT_USERNAME, NARRATOR_NAME
from context import Context
from contrib.lm_prefix_utils import get_character_priming
from decorators import monitor
from exceptions import MissingContextData
from structured_prefix import StructuredPrefix
from template import TemplatingEngine


# TODO: use ansible core filters instead of custom filters.
# BEGIN: Hermes Filters
# TODO: necessary for correct YAML parsing. Figure out a better solution.
# Tried using folded block scalar ">" but it didn't work.
def escape_sequences(message: str) -> str:
    """Escape sequences that will break yaml parsing."""
    return message.replace(Context.NEWLINE, Context.ESCAPED_NEWLINE).replace(
        Context.CARRIAGE_RETURN, Context.ESCAPED_CARRIAGE_RETURN
    )


def maybe_inject_narrator(message: str, default_author: str = NARRATOR_NAME) -> str:
    """Inject narrator into the message if applicable."""
    if re.match(r"^[\w-]+:", message):
        return message
    return f"{default_author}: {message}"


def raise_missing_context_data(key: str):
    """Raise missing data from jinja context."""
    raise MissingContextData(f"Missing required key in jinja context: {key=}")


def canonicalize_user_name(name: Union[str, None]) -> str:
    """Makes name format consistent with author names we use in training data."""
    # The "-" is used in upstream components and should be overriden to default value.
    if name is None or not name or name == "-":
        return DEFAULT_USERNAME
    return "-".join(name.split())


# END: Hermes Filters


@monitor
def build_structured_prefix(
    contextual_logger: LoggerAdapter,
    structured_prefix: StructuredPrefix,
    close_last_message: bool = False,
    truncation_step: int = None,
) -> dict:
    """Build structured prefix using Hermes templating engine."""
    del close_last_message  # TODO: support long streaming.

    # TODO: move templating engine into Context class.
    engine = TemplatingEngine(logger=contextual_logger)
    jinja_context = _build_jinja_context(
        structured_prefix=structured_prefix,
        contextual_logger=contextual_logger,
    )

    rendered_template = engine.render_template(
        jinja_context,
        template_name=structured_prefix.hermes_generation_template_name,
        raw_template=structured_prefix.hermes_generation_template,
    )

    # TODO: add prometheus logging for latency.
    lm_context = Context(
        contextual_logger=contextual_logger,
        rendered_template=rendered_template,
        structured_prefix=structured_prefix,
        truncation_step=truncation_step,
    )
    lm_context.load_yaml()
    lm_context.tokenize()
    lm_context.truncate()
    lm_context.validate()

    return {
        "character_definitions": structured_prefix.character_definitions,
        "chat_history": structured_prefix.chat_history[
            -lm_context.num_messages_included :
        ],
        "chat_hist_global_index": 0,
        "reply_prompt": lm_context.raw_reply_prompt,
        "space_added": True,
        "token_limit": structured_prefix.token_limit,
        "tokenized_context": lm_context.tokenized_context,
        "timestamp": lm_context.timestamp_str,
        "lm_context": lm_context,
    }


@monitor
def _build_jinja_context(
    structured_prefix: StructuredPrefix,
    contextual_logger: LoggerAdapter,
) -> dict:
    if structured_prefix.raw_prompt_data_dict is None:
        structured_prefix.parse_raw_prompt_data(contextual_logger=contextual_logger)
    if structured_prefix.chat_context_messages is None:
        structured_prefix.parse_chat_context_messages(
            contextual_logger=contextual_logger
        )

    # TODO: handle this in the template not in logical layer.
    username = structured_prefix.raw_prompt_data_dict.get("username", "")
    if not username or username == "-":
        username = DEFAULT_USERNAME

    return {
        ### Raw prompt data that has not been tampered with ###
        "chat_type": structured_prefix.raw_prompt_data_dict.get("chat_type", ""),
        "character": structured_prefix.raw_prompt_data_dict.get("character", {}),
        "user_country_code": structured_prefix.raw_prompt_data_dict.get(
            "user_country_code", ""
        ),
        "username": username,
        "persona_definition": structured_prefix.raw_prompt_data_dict.get(
            "persona_definition", ""
        ),
        "is_proactive": structured_prefix.raw_prompt_data_dict.get(
            "is_proactive", False
        ),
        "proactive_metadata": structured_prefix.raw_prompt_data_dict.get(
            "proactive_metadata", {}
        ),
        "chat_context_messages": structured_prefix.chat_context_messages,
        ###  Constants data ###
        "narrator_name": NARRATOR_NAME,
        ### Legacy prompt data that has been tampered with by upstream components ###
        "character_definition_messages": structured_prefix.character_definitions,
        "pinned_history": structured_prefix.pinned_history or [],
        "chat_history": structured_prefix.chat_history,
        "reply_prompt": structured_prefix.reply_prompt,
        "timestamp": structured_prefix.timestamp,
        ### Filters ###
        # TODO: import ansible core filters instead of custom filters.
        "maybe_inject_narrator": maybe_inject_narrator,
        "escape_sequences": escape_sequences,
        "get_character_priming": get_character_priming,
        "canonicalize_user_name": canonicalize_user_name,
        "raise_missing_context_data": raise_missing_context_data,
    }



"""Templating engine for rendering Jinja templates."""

import logging
from importlib import resources

from cachetools import TTLCache
from decorators import monitor
from exceptions import TemplateNotFoundError
from jinja2 import Environment, FileSystemLoader, Template
from jinja2.exceptions import TemplateNotFound
from metrics import Metrics

CACHE_MAX_SIZE = 100
CACHE_TTL_SECS = 30


class TemplatingEngine:
    """Template engine."""

    _instance = None
    _template_cache = TTLCache(maxsize=CACHE_MAX_SIZE, ttl=CACHE_TTL_SECS)

    def __new__(cls, *args, **kwargs):
        """Singleton pattern that allows arbitrary arguments."""
        if cls._instance is None:
            cls._instance = super(TemplatingEngine, cls).__new__(cls)
            # Initialize _instance's attributes
            cls._instance._initialized = False

        # Replace the contextual logger if provided.
        cls._instance._logger = (
            kwargs.pop("logger", None)
            if "logger" in kwargs
            else logging.getLogger(__name__)
        )
        return cls._instance

    def __init__(
        self, logger: logging.LoggerAdapter = None, use_template_cache: bool = True
    ):
        """Initialize template engine."""
        if not self._initialized:
            self._default_template = None
            self._logger = logger if logger else logging.getLogger(__name__)
            self._use_template_cache = use_template_cache
            self._initialized = True

    @monitor
    def render_template(
        self, context: dict = None, raw_template: str = None, template_name: str = None
    ):
        """Render the jinja template with the provided context data."""
        if context is None:
            context = {}

        if raw_template and template_name:
            raise ValueError("Cannot provide both raw_template and template_name.")

        # Template was provided at runtime.
        if raw_template:
            self._logger.info(
                f"### Hermes is using a runtime template: \n\n {raw_template=}"
            )

            tmp_template = self._load_template_from_string(raw_template)
            return tmp_template.render(context)

        # Load template from disk every time.
        if template_name:
            try:
                tmp_template = self._get_template(template_name=template_name)
                return tmp_template.render(context)
            except TemplateNotFound as ex:
                raise TemplateNotFoundError(
                    f"Template not found: {template_name}. {ex=}"
                )

        # If no template is provided or template could not be rendered, fallback to the default.
        if self._default_template is None:
            self._default_template = self._get_template()

        return self._default_template.render(context)

    def _get_template(self, template_name: str = "production.yml.j2") -> Template:
        """Get template from cache or load from disk."""
        if not self._use_template_cache or template_name not in self._template_cache:
            self._template_cache[template_name] = self._load_template(template_name)
            Metrics().HERMES_TEMPLATE_NAME.labels(
                template_name=template_name, source="disk"
            ).inc()
        else:
            Metrics().HERMES_TEMPLATE_NAME.labels(
                template_name=template_name, source="cache"
            ).inc()

        return self._template_cache[template_name]

    def _load_template(self, template_name: str) -> Template:
        """Load template from disk."""
        try:
            multiplexed_path = resources.files("hermes_cai.templates")
            # HACK: This is a brittle hack to get the path to the templates.
            # Replace when we decouple the registry.
            pkg_resource_path = str(multiplexed_path).split("'")[1]
        except ModuleNotFoundError:
            self._logger.error(
                "Could not find the templates directory. Falling back to local."
            )
            pkg_resource_path = "./templates"
        loader = FileSystemLoader(searchpath=pkg_resource_path)
        env = Environment(loader=loader)
        # TODO: support dynamic routing to different templates.
        template = env.get_template(template_name)
        return template

    def _load_template_from_string(self, raw_template: str):
        """Load template from raw string."""
        return Template(raw_template)



"""Context module for Hermes CAI."""

import math
import re
from dataclasses import dataclass
from functools import reduce
from logging import LoggerAdapter

import yaml
from constants import VALID_PROMPT_STRING_PATTERNS
from contrib.lm_prefix_utils import TokenizedContext, get_tokenizer
from decorators import monitor
from exceptions import (
    InvalidPromptError,
    MessageStartIdxNotFound,
    TimestampIdxNotFoundError,
    TimestampStrNotFoundError,
    TokenLimitExceededError,
)
from metrics import Metrics
from structured_prefix import StructuredPrefix
from yaml import CSafeLoader


@dataclass
class PromptPart:
    """Container representing repeated prompt parts from the template."""

    name: str
    raw_string: str
    tokens: list[int] = None


class Context:
    """Context holds all essential data and business logic for constructing prompts."""

    BOD_PART_NAME: str = "bod_ts"
    REPLY_PROMPT_PART_NAME: str = "reply_prompt"
    CACHE_FRIENDLY_TRUNCATION_STEP: int = 1000
    BOD: str = "<|beginningofdialog|>"
    BOM: str = "<|beginningofmessage|>"
    SPACE_MARKER: str = "<|space|>"
    CARRIAGE_RETURN: str = "\r"
    ESCAPED_CARRIAGE_RETURN: str = "\\r"
    NEWLINE: str = "\n"
    ESCAPED_NEWLINE: str = "\\n"
    SINGLE_QUOTE: str = "'"
    ESCAPED_SINGLE_QUOTE: str = "'"

    @monitor
    def __init__(
        self,
        contextual_logger: LoggerAdapter,
        rendered_template: str,
        structured_prefix: StructuredPrefix,
        truncation_step: int = None,
    ):
        """Initialize context."""
        if truncation_step is not None and truncation_step <= 0:
            raise ValueError(f"Invalid truncation step: {truncation_step=}")

        self._logger = contextual_logger
        self.prompt_parts = None
        # Special case for reconstructing full prompts that have been truncated.
        self.prompt_parts_bak = None
        self.rendered_template = rendered_template
        self.token_limit = structured_prefix.token_limit
        self.tokenizer = get_tokenizer()
        self.idx_after_timestamp = -1
        self.timestamp_str = None
        self.num_total_pinned_message_tokens = 0
        self.num_total_tokens = 0
        self.num_tokens_truncated = 0
        self.message_start_idx = -1
        self.num_messages_included = 0
        self.num_messages_truncated = 0
        self.num_total_pinned_messages = 0
        self.num_total_messages = 0
        self.raw_reply_prompt = ""
        self.truncation_step = truncation_step

    @monitor
    def load_yaml(self):
        """Load yaml from rendered template."""
        self.prompt_parts = self._load_yaml(self.rendered_template)

    @monitor
    def validate(self):
        """Raises if the current Context object is invalid by certain heuristics."""
        # Cache prompt tokens and prompt string as it can be expensive to compute.
        prompt_tokens = self.prompt_tokens
        prompt_string = self.prompt_string

        if len(self.prompt_parts) > 1 and self.prompt_parts[1].raw_string.startswith(
            Context.BOM
        ):
            raise InvalidPromptError(
                f"First message should NOT contain {Context.BOM=}: {self.prompt_parts[1]=}"
            )

        for pattern in VALID_PROMPT_STRING_PATTERNS:
            if not re.match(pattern, prompt_string):
                raise InvalidPromptError(
                    f"Invalid prompt regex: {pattern=} {str(prompt_string)=}"
                )

        if len(prompt_tokens) > self.token_limit:
            raise TokenLimitExceededError(
                f"Token limit exceeded: {len(prompt_tokens)=} {self.token_limit=}"
            )

        if not self.timestamp_str:
            raise TimestampStrNotFoundError(
                f"Timestamp string not found: {self.timestamp_str=}"
            )

        if self.idx_after_timestamp == -1:
            raise TimestampIdxNotFoundError(
                f"Timestamp index not found: {self.idx_after_timestamp=}"
            )

    @monitor
    def tokenize(self):
        """Cleans, tokenizes and maybe extracts some metadata of each prompt part."""
        # Tokenize each prompt part.
        for idx, part in enumerate(self.prompt_parts):
            self._clean_raw_string(part)
            self._tokenize_part(part)
            self._maybe_extract_metadata_and_clean(idx=idx, prompt_part=part)

    @monitor
    def truncate(self):
        """Truncate messages if the token limit is exceeded."""
        if self.message_start_idx == -1:
            raise MessageStartIdxNotFound(
                f"Message start index not found: {self.message_start_idx=}"
            )

        num_tokens_to_truncate = self._calculate_num_tokens_to_truncate()
        i = self.message_start_idx
        if num_tokens_to_truncate:
            self._logger.info(
                f"Truncating... {num_tokens_to_truncate=}; {self.num_total_tokens=}; {self.num_total_messages=}"
            )
        while (
            self.num_tokens_truncated < num_tokens_to_truncate
            and i < self.message_start_idx + self.num_total_messages
        ):
            part = self.prompt_parts[i]
            self.num_tokens_truncated += len(part.tokens)
            self.num_messages_truncated += 1
            i += 1

        self.num_messages_included = (
            self.num_total_messages - self.num_messages_truncated
        )
        self._truncate()

        # A short-term hack to fix the special case of first message should not have bom.
        # TODO: Rethink this in the future to see if we can take it out of the logical layer.
        if (
            len(self.prompt_parts) > 1
            and Context.BOM in self.prompt_parts[1].raw_string
        ):
            self.prompt_parts[1].raw_string = self.prompt_parts[1].raw_string.replace(
                Context.BOM, " "
            )
            self._tokenize_part(self.prompt_parts[1])

    @property
    def tokenized_context(self) -> TokenizedContext:
        """Tokenized context of self at a point in time."""
        return TokenizedContext(
            tokens=self.prompt_tokens,
            idx_after_timestamp=self.idx_after_timestamp,
            context_num_msg=self.num_messages_included,
            truncated_num_msg=self.num_messages_truncated,
        )

    @property
    def prompt_tokens(self) -> list[int]:
        """Raw promp tokens at a point in time."""
        return reduce(lambda acc, part: acc + part.tokens, self.prompt_parts, [])

    @property
    def prompt_string(self) -> str:
        """Raw prompt string at a point in time."""
        return self.tokenizer.detokenize(self.prompt_tokens)

    @property
    def prompt_string_pretruncation(self) -> str:
        """Raw prompt string before truncation."""
        return reduce(
            lambda acc, part: acc + part.raw_string, self.prompt_parts_bak, ""
        )

    @property
    def total_active_tokens(self) -> int:
        """Total number of tokens in the context."""
        return self.num_total_tokens - self.num_tokens_truncated

    @monitor
    def _load_yaml(self, rendered_template: str) -> list[PromptPart]:
        """Load yaml from rendered template."""
        try:
            loaded_yaml = yaml.load(rendered_template, Loader=CSafeLoader)
        except Exception as ex:
            self._logger.error(
                f"### Hermes: Error loading yaml from rendered template: {rendered_template=}; {ex=}",
                exc_info=True,
            )
            raise ex

        return list(
            map(
                lambda x: PromptPart(**x),
                loaded_yaml,
            )
        )

    def _calculate_num_tokens_to_truncate(self):
        """Calculates the number of messages to truncate."""
        if self.truncation_step is None:
            self.truncation_step = Context.CACHE_FRIENDLY_TRUNCATION_STEP

        num_surplus_tokens = max(0, self.num_total_tokens - self.token_limit)
        return (
            math.ceil(num_surplus_tokens / self.truncation_step) * self.truncation_step
        )

    def _truncate(self):
        """Truncate messages from prompt_parts; modifies prompt_parts in place."""
        self._observe_context_metrics()
        if self.message_start_idx == -1:
            raise MessageStartIdxNotFound(
                f"Message start index not found: {self.message_start_idx=}"
            )

        if self.num_messages_truncated == 0:
            self._logger.info("No truncation needed.")
            return

        if self.num_messages_truncated == self.num_total_messages:
            self._logger.warn("Truncating all messages!")

        self.prompt_parts_bak = self.prompt_parts
        self.prompt_parts = (
            self.prompt_parts[: self.message_start_idx]
            + self.prompt_parts[self.message_start_idx + self.num_messages_truncated :]
        )

    # TODO: move this out of the wheel.
    def _observe_context_metrics(self):
        """Observe truncation metrics."""
        # TODO: use constants.
        Metrics().TOKEN_METRICS.labels(metric_type="total_tokens").observe(
            self.num_total_tokens
        )
        Metrics().TOKEN_METRICS.labels(metric_type="truncated_tokens").observe(
            self.num_tokens_truncated
        )
        Metrics().TOKEN_METRICS.labels(metric_type="active_tokens").observe(
            self.total_active_tokens
        )
        if self.num_total_pinned_message_tokens > 0:
            Metrics().TOKEN_METRICS.labels(
                metric_type="total_pinned_message_tokens"
            ).observe(self.num_total_pinned_message_tokens)

        Metrics().MESSAGE_METRICS.labels(metric_type="total_rendered_messages").observe(
            self.num_total_messages
        )
        Metrics().MESSAGE_METRICS.labels(metric_type="truncated_messages").observe(
            self.num_messages_truncated
        )
        Metrics().MESSAGE_METRICS.labels(metric_type="total_active_messages").observe(
            self.num_messages_included
        )

        if self.num_total_pinned_messages > 0:
            Metrics().MESSAGE_METRICS.labels(
                metric_type="truncated_messages_with_pinned"
            ).observe(self.num_messages_truncated)
            Metrics().MESSAGE_METRICS.labels(
                metric_type="total_active_messages_with_pinned"
            ).observe(self.num_messages_included)

    def _tokenize_part(self, prompt_part: PromptPart):
        prompt_part.tokens = self.tokenizer.tokenize(prompt_part.raw_string)
        self.num_total_tokens += len(prompt_part.tokens)
        if prompt_part.name.startswith("pinned_message"):
            self.num_total_pinned_message_tokens += len(prompt_part.tokens)

    def _clean_raw_string(self, prompt_part: PromptPart):
        # Cleanup whitespace, newlines and add whitespace where it is explicitly marked.
        prompt_part.raw_string = (
            prompt_part.raw_string.strip()
            .replace(Context.SPACE_MARKER, " ")
            .replace(Context.ESCAPED_NEWLINE, Context.NEWLINE)
            .replace(Context.ESCAPED_CARRIAGE_RETURN, Context.CARRIAGE_RETURN)
            .replace(Context.ESCAPED_SINGLE_QUOTE, Context.SINGLE_QUOTE)
        )

    def _maybe_extract_metadata_and_clean(self, idx: int, prompt_part: PromptPart):
        if prompt_part.name == Context.BOD_PART_NAME:
            self.idx_after_timestamp = len(prompt_part.tokens) + 1
            self.timestamp_str = prompt_part.raw_string.replace(Context.BOD, "")

        if prompt_part.name == Context.REPLY_PROMPT_PART_NAME:
            self.raw_reply_prompt = prompt_part.raw_string.replace(Context.BOM, "")

        if prompt_part.name.startswith("message"):
            if self.message_start_idx == -1:
                self.message_start_idx = idx
            self.num_total_messages += 1

        if prompt_part.name.startswith("pinned_message"):
            self.num_total_pinned_messages += 1



"""Setup file for the Hermes package."""

import os

from setuptools import find_packages, setup

# Get the current directory of the setup.py script
here = os.path.abspath(os.path.dirname(__file__))

# Read the version from the VERSION file
with open(os.path.join(here, "VERSION")) as version_file:
    version = version_file.read().strip()

setup(
    name="hermes-cai",
    version=version,
    packages=find_packages(include=["hermes_cai", "hermes_cai.*"]),
    include_package_data=True,
    package_data={"hermes_cai": ["templates/*", "contrib/vocab/*"]},
    install_requires=[],
    author="James Groeneveld",
    author_email="james@character.ai",
    description="Defining and constructing production-grade LLM prompts via rich structured templates.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/character-tech/chat-stack",
    python_requires=">=3.10",
    license="MIT",
)



"""Custom exceptions for the Hermes engine."""


class InvalidPromptError(Exception):
    """Invalid prompt error."""


class TokenLimitExceededError(Exception):
    """Token limit exceeded error."""


class TimestampStrNotFoundError(Exception):
    """Timestamp string not found error."""


class TimestampIdxNotFoundError(Exception):
    """Timestamp index not found error."""


class MessageStartIdxNotFound(Exception):
    """Message start index not found error."""


class MissingContextData(Exception):
    """Missing context data error."""


class TemplateNotFoundError(Exception):
    """Template not found error."""
